# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
:mod:`vsm` -- Cloud IaaS Platform
===================================

.. automodule:: vsm
   :platform: Unix
   :synopsis: Infrastructure-as-a-Service Cloud platform.
.. moduleauthor:: Jesse Andrews <jesse@ansolabs.com>
.. moduleauthor:: Devin Carlen <devin.carlen@gmail.com>
.. moduleauthor:: Vishvananda Ishaya <vishvananda@gmail.com>
.. moduleauthor:: Joshua McKenty <joshua@cognition.ca>
.. moduleauthor:: Manish Singh <yosh@gimp.org>
.. moduleauthor:: Andy Smith <andy@anarkystic.com>
"""

import gettext
import sys
import pkg_resources

# If there is a conflicting non egg module,
# i.e. an older standard system module installed,
# then replace it with this requirement
def replace_dist(requirement):
    try:
        return pkg_resources.require(requirement)
    except pkg_resources.VersionConflict:
        e = sys.exc_info()[1]
        dist=e.args[0]
        req=e.args[1]
        if dist.key == req.key and not dist.location.endswith('.egg'):
            del pkg_resources.working_set.by_key[dist.key]
            # We assume there is no need to adjust sys.path
            # and the associated pkg_resources.working_set.entries
            return pkg_resources.require(requirement)

replace_dist("WebOb >= 1.0")
replace_dist("SQLAlchemy >= 0.6.3")
replace_dist("Routes >= 1.12.3")

replace_dist("PasteDeploy >= 1.5.0")
# This hack is needed because replace_dist() results in
# the standard paste module path being at the start of __path__.
# TODO: See can we get pkg_resources to do the right thing directly
import paste
paste.__path__.insert(0, paste.__path__.pop(-1))
gettext.install('vsm', unicode=1)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 Intel
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
App nodes operations.
"""

from vsm import context
from vsm import db
from vsm import utils
from vsm import exception
from vsm.openstack.common.gettextutils import _
from vsm.openstack.common import log as logging
from vsm.openstack.common.db import exception as db_exc

import json
import urllib2


LOG = logging.getLogger(__name__)

def get_all_nodes(contxt):
    """get all non-deletd app nodes as a dict"""
    if contxt is None:
        contxt = context.get_admin_context()
    try:
        nodes = db.appnodes_get_all(contxt)
        return nodes
    except db_exc.DBError as e:
        LOG.exception(_("DB Error on getting Appnodes %s" % e))
        raise exception.AppNodeFailure()

def _check_v3(tenant_name, username, password, auth_url, region_name):
    result = False
    auth_url = auth_url + "/auth/tokens"
    user_id = username
    user_password = password
    project_id = tenant_name
    auth_data = {
        "auth": {
            "identity": {
                "methods": ["password"],
                "password": {
                    "user": {
                        "id": user_id,
                        "password": user_password
                    }
                }
            },
            "scope": {
                "project": {
                    "id": project_id
                }
            }
        }
    }
    auth_request = urllib2.Request(auth_url)
    auth_request.add_header("content-type", "application/json")
    auth_request.add_header('Accept', 'application/json')
    auth_request.add_header('User-Agent', 'python-mikeyp')
    auth_request.add_data(json.dumps(auth_data))
    auth_response = urllib2.urlopen(auth_request)
    response_data = json.loads(auth_response.read())
    services_list = response_data['token']['catalog']
    for service in services_list:
        service_type = service['type']
        service_name = service['name']
        endpoints_list = service['endpoints']
        if service_type == "volume" and service_name == "cinder":
            region_list = []
            for endpoint in endpoints_list:
                region = endpoint['region']
                if region not in region_list:
                    region_list.append(region)
                if region_name != None and region == region_name:
                    result = True
                    break
            if len(region_list) == 1:
                result = True
                break
    return result

def _check_v2(tenant_name, username, password, auth_url, region_name):
    result = False
    auth_url = auth_url + "/tokens"
    auth_data = {
        "auth": {
            "tenantName": tenant_name,
            "passwordCredentials": {
                "username": username,
                "password": password
            }
        }
    }
    auth_request = urllib2.Request(auth_url)
    auth_request.add_header("content-type", "application/json")
    auth_request.add_header('Accept', 'application/json')
    auth_request.add_header('User-Agent', 'python-mikeyp')
    auth_request.add_data(json.dumps(auth_data))
    auth_response = urllib2.urlopen(auth_request)
    response_data = json.loads(auth_response.read())
    service_list = response_data['access']['serviceCatalog']
    for service in service_list:
        if service['type'] == 'volume' and service['name'] == 'cinder':
            for endpoint in service['endpoints']:
                if region_name != None and endpoint['region'] == region_name:
                    result = True
                    break
                elif region_name == None or region_name == "" and len(service['endpoints']) == 1:
                    result = True
                    break
    return result

def create(contxt, auth_openstack=None, allow_duplicate=False):
    """create app node from a dict"""
    if contxt is None:
        contxt = context.get_admin_context()

    appnode = auth_openstack
    if not appnode:
        raise exception.AppNodeInvalidInfo()

    auth_url = appnode['os_auth_url'].strip("/")
    ssh_user = appnode['ssh_user']
    tenant_name = appnode['os_tenant_name']
    username = appnode['os_username']
    password = appnode['os_password']
    region_name = appnode['os_region_name']

    os_controller_host = auth_url.split(":")[1][2:]
    result, err = utils.execute(
        'check_xtrust_crudini',
        ssh_user,
        os_controller_host,
        run_as_root = True
    )
    LOG.info("==============result: %s" % result)
    LOG.info("==============err: %s" % err)
    if "command not found" in err:
        raise Exception("Command not found on %s" % os_controller_host)
    if "Permission denied" in err:
        raise Exception("Please check the mutual trust between vsm controller node "
                        "and openstack controller node")
    if "No passwd entry" in err:
        raise Exception("Please check the trust user")

    # support keystone v3 ad v2.0
    keystone_version = auth_url.split("/")[-1]
    try:
        if keystone_version == "v3":
            result = _check_v3(tenant_name,
                               username,
                               password,
                               auth_url,
                               region_name)
        elif keystone_version == "v2.0":
            result = _check_v2(tenant_name,
                               username,
                               password,
                               auth_url,
                               region_name)
        else:
            raise Exception("Only support keystone v3 and v2.0 now.")
        if result:
            appnode['ssh_status'] = "reachable"
        else:
            appnode['ssh_status'] = "no cinder-volume"
    except:
        LOG.exception(_("Error to access to openstack"))
        appnode['ssh_status'] = "unreachable"

    ref = []
    try:
        ref.append(db.appnodes_create(contxt, appnode, allow_duplicate))
    except db_exc.DBError as e:
        LOG.exception(_("DB Error on creating Appnodes %s" % e))
        raise exception.AppNodeFailure()
    return ref

def update(contxt, appnode_id, appnode):
    """update app node ssh status, log info or deleted"""
    if contxt is None:
        contxt = context.get_admin_context()

    auth_url = appnode['os_auth_url'].strip("/")
    ssh_user = appnode['ssh_user']
    tenant_name = appnode['os_tenant_name']
    username = appnode['os_username']
    password = appnode['os_password']
    region_name = appnode['os_region_name']

    id = utils.int_from_str(appnode_id)
    LOG.debug('app node id: %s ' % id)

    os_controller_host = auth_url.split(":")[1][2:]
    result, err = utils.execute(
            'check_xtrust_crudini',
            ssh_user,
            os_controller_host,
            run_as_root = True
    )
    LOG.info("==============result: %s" % result)
    LOG.info("==============err: %s" % err)
    if "command not found" in err:
        raise Exception("Command not found on %s" % os_controller_host)
    if "Permission denied" in err:
        raise Exception("Please check the mutual trust between vsm controller node "
                        "and openstack controller node")
    if "No passwd entry" in err:
        raise Exception("Please check the trust user")

    # support keystone v3 ad v2.0
    keystone_version = auth_url.split("/")[-1]
    try:
        if keystone_version == "v3":
            result = _check_v3(tenant_name,
                               username,
                               password,
                               auth_url,
                               region_name)
        elif keystone_version == "v2.0":
            result = _check_v2(tenant_name,
                               username,
                               password,
                               auth_url,
                               region_name)
        else:
            raise Exception("Only support keystone v3 and v2.0 now.")
        if result:
            appnode['ssh_status'] = "reachable"
        else:
            appnode['ssh_status'] = "no cinder-volume"
    except:
        LOG.exception(_("Error to access to openstack"))
        appnode['ssh_status'] = "unreachable"

    try:
        return db.appnodes_update(contxt, id, appnode)
    except db_exc.DBError as e:
        LOG.exception(_("DB Error on updating Appnodes %s" % e))
        raise exception.AppNodeFailure()

def destroy(contxt, appnode_id):
    if contxt is None:
        contxt = context.get_admin_context()

    appnode_id = utils.int_from_str(appnode_id)
    try:
        db.appnodes_destroy(contxt, appnode_id)
    except db_exc.DBError as e:
        LOG.exception(_("DB Error on deleting Appnodes %s" % e))
        raise exception.AppNodeFailure()



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2010 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Implementation of an image service that uses Glance as the backend"""

from __future__ import absolute_import

import copy
import itertools
import random
import sys
import time
import urlparse

import glanceclient
import glanceclient.exc

from vsm import exception
from vsm import flags
from vsm.openstack.common import jsonutils
from vsm.openstack.common import log as logging
from vsm.openstack.common import timeutils

LOG = logging.getLogger(__name__)
FLAGS = flags.FLAGS

def _parse_image_ref(image_href):
    """Parse an image href into composite parts.

    :param image_href: href of an image
    :returns: a tuple of the form (image_id, host, port)
    :raises ValueError

    """
    url = urlparse.urlparse(image_href)
    port = url.port or 80
    host = url.netloc.split(':', 1)[0]
    image_id = url.path.split('/')[-1]
    use_ssl = (url.scheme == 'https')
    return (image_id, host, port, use_ssl)

def _create_glance_client(context, host, port, use_ssl,
                          version=FLAGS.glance_api_version):
    """Instantiate a new glanceclient.Client object"""
    if version is None:
        version = FLAGS.glance_api_version
    if use_ssl:
        scheme = 'https'
    else:
        scheme = 'http'
    params = {}
    params['insecure'] = FLAGS.glance_api_insecure
    if FLAGS.auth_strategy == 'keystone':
        params['token'] = context.auth_token
    endpoint = '%s://%s:%s' % (scheme, host, port)
    return glanceclient.Client(str(version), endpoint, **params)

def get_api_servers():
    """
    Shuffle a list of FLAGS.glance_api_servers and return an iterator
    that will cycle through the list, looping around to the beginning
    if necessary.
    """
    api_servers = []
    for api_server in FLAGS.glance_api_servers:
        if '//' not in api_server:
            api_server = 'http://' + api_server
        url = urlparse.urlparse(api_server)
        port = url.port or 80
        host = url.netloc.split(':', 1)[0]
        use_ssl = (url.scheme == 'https')
        api_servers.append((host, port, use_ssl))
    random.shuffle(api_servers)
    return itertools.cycle(api_servers)

class GlanceClientWrapper(object):
    """Glance client wrapper class that implements retries."""

    def __init__(self, context=None, host=None, port=None, use_ssl=False,
                 version=None):
        if host is not None:
            self.client = self._create_static_client(context,
                                                     host, port,
                                                     use_ssl, version)
        else:
            self.client = None
        self.api_servers = None
        self.version = version

    def _create_static_client(self, context, host, port, use_ssl, version):
        """Create a client that we'll use for every call."""
        self.host = host
        self.port = port
        self.use_ssl = use_ssl
        self.version = version
        return _create_glance_client(context,
                                     self.host, self.port,
                                     self.use_ssl, self.version)

    def _create_onetime_client(self, context, version):
        """Create a client that will be used for one call."""
        if self.api_servers is None:
            self.api_servers = get_api_servers()
        self.host, self.port, self.use_ssl = self.api_servers.next()
        return _create_glance_client(context,
                                     self.host, self.port,
                                     self.use_ssl, version)

    def call(self, context, method, *args, **kwargs):
        """
        Call a glance client method.  If we get a connection error,
        retry the request according to FLAGS.glance_num_retries.
        """
        version = self.version
        if version in kwargs:
            version = kwargs['version']

        retry_excs = (glanceclient.exc.ServiceUnavailable,
                      glanceclient.exc.InvalidEndpoint,
                      glanceclient.exc.CommunicationError)
        num_attempts = 1 + FLAGS.glance_num_retries

        for attempt in xrange(1, num_attempts + 1):
            client = self.client or self._create_onetime_client(context,
                                                                version)
            try:
                return getattr(client.images, method)(*args, **kwargs)
            except retry_excs as e:
                host = self.host
                port = self.port
                extra = "retrying"
                error_msg = _("Error contacting glance server "
                              "'%(host)s:%(port)s' for '%(method)s', "
                              "%(extra)s.")
                if attempt == num_attempts:
                    extra = 'done trying'
                    LOG.exception(error_msg, locals())
                    raise exception.GlanceConnectionFailed(host=host,
                                                           port=port,
                                                           reason=str(e))
                LOG.exception(error_msg, locals())
                time.sleep(1)

class GlanceImageService(object):
    """Provides storage and retrieval of disk image objects within Glance."""

    def __init__(self, client=None):
        self._client = client or GlanceClientWrapper()

    def detail(self, context, **kwargs):
        """Calls out to Glance for a list of detailed image information."""
        params = self._extract_query_params(kwargs)
        try:
            images = self._client.call(context, 'list', **params)
        except Exception:
            _reraise_translated_exception()

        _images = []
        for image in images:
            if self._is_image_available(context, image):
                _images.append(self._translate_from_glance(image))

        return _images

    def _extract_query_params(self, params):
        _params = {}
        accepted_params = ('filters', 'marker', 'limit',
                           'sort_key', 'sort_dir')
        for param in accepted_params:
            if param in params:
                _params[param] = params.get(param)

        # ensure filters is a dict
        _params.setdefault('filters', {})
        # NOTE(vish): don't filter out private images
        _params['filters'].setdefault('is_public', 'none')

        return _params

    def show(self, context, image_id):
        """Returns a dict with image data for the given opaque image id."""
        try:
            image = self._client.call(context, 'get', image_id)
        except Exception:
            _reraise_translated_image_exception(image_id)

        if not self._is_image_available(context, image):
            raise exception.ImageNotFound(image_id=image_id)

        base_image_meta = self._translate_from_glance(image)
        return base_image_meta

    def get_location(self, context, image_id):
        """Returns the direct url representing the backend storage location,
        or None if this attribute is not shown by Glance."""
        try:
            client = GlanceClientWrapper()
            image_meta = client.call(context, 'get', image_id)
        except Exception:
            _reraise_translated_image_exception(image_id)

        if not self._is_image_available(context, image_meta):
            raise exception.ImageNotFound(image_id=image_id)

        return getattr(image_meta, 'direct_url', None)

    def download(self, context, image_id, data):
        """Calls out to Glance for metadata and data and writes data."""
        try:
            image_chunks = self._client.call(context, 'data', image_id)
        except Exception:
            _reraise_translated_image_exception(image_id)

        for chunk in image_chunks:
            data.write(chunk)

    def create(self, context, image_meta, data=None):
        """Store the image data and return the new image object."""
        sent_service_image_meta = self._translate_to_glance(image_meta)

        if data:
            sent_service_image_meta['data'] = data

        recv_service_image_meta = self._client.call(context, 'create',
                                                    **sent_service_image_meta)

        return self._translate_from_glance(recv_service_image_meta)

    def update(self, context, image_id,
               image_meta, data=None, purge_props=True):
        """Modify the given image with the new data."""
        image_meta = self._translate_to_glance(image_meta)
        image_meta['purge_props'] = purge_props
        #NOTE(bcwaldon): id is not an editable field, but it is likely to be
        # passed in by calling code. Let's be nice and ignore it.
        image_meta.pop('id', None)
        if data:
            image_meta['data'] = data
        try:
            image_meta = self._client.call(context, 'update', image_id,
                                           **image_meta)
        except Exception:
            _reraise_translated_image_exception(image_id)
        else:
            return self._translate_from_glance(image_meta)

    def delete(self, context, image_id):
        """Delete the given image.

        :raises: ImageNotFound if the image does not exist.
        :raises: NotAuthorized if the user is not an owner.

        """
        try:
            self._client.call(context, 'delete', image_id)
        except glanceclient.exc.NotFound:
            raise exception.ImageNotFound(image_id=image_id)
        return True

    @staticmethod
    def _translate_to_glance(image_meta):
        image_meta = _convert_to_string(image_meta)
        image_meta = _remove_read_only(image_meta)
        return image_meta

    @staticmethod
    def _translate_from_glance(image):
        image_meta = _extract_attributes(image)
        image_meta = _convert_timestamps_to_datetimes(image_meta)
        image_meta = _convert_from_string(image_meta)
        return image_meta

    @staticmethod
    def _is_image_available(context, image):
        """Check image availability.

        This check is needed in case Nova and Glance are deployed
        without authentication turned on.
        """
        # The presence of an auth token implies this is an authenticated
        # request and we need not handle the noauth use-case.
        if hasattr(context, 'auth_token') and context.auth_token:
            return True

        if image.is_public or context.is_admin:
            return True

        properties = image.properties

        if context.project_id and ('owner_id' in properties):
            return str(properties['owner_id']) == str(context.project_id)

        if context.project_id and ('project_id' in properties):
            return str(properties['project_id']) == str(context.project_id)

        try:
            user_id = properties['user_id']
        except KeyError:
            return False

        return str(user_id) == str(context.user_id)

def _convert_timestamps_to_datetimes(image_meta):
    """Returns image with timestamp fields converted to datetime objects."""
    for attr in ['created_at', 'updated_at', 'deleted_at']:
        if image_meta.get(attr):
            image_meta[attr] = timeutils.parse_isotime(image_meta[attr])
    return image_meta

# NOTE(bcwaldon): used to store non-string data in glance metadata
def _json_loads(properties, attr):
    prop = properties[attr]
    if isinstance(prop, basestring):
        properties[attr] = jsonutils.loads(prop)

def _json_dumps(properties, attr):
    prop = properties[attr]
    if not isinstance(prop, basestring):
        properties[attr] = jsonutils.dumps(prop)

_CONVERT_PROPS = ('block_device_mapping', 'mappings')

def _convert(method, metadata):
    metadata = copy.deepcopy(metadata)
    properties = metadata.get('properties')
    if properties:
        for attr in _CONVERT_PROPS:
            if attr in properties:
                method(properties, attr)

    return metadata

def _convert_from_string(metadata):
    return _convert(_json_loads, metadata)

def _convert_to_string(metadata):
    return _convert(_json_dumps, metadata)

def _extract_attributes(image):
    IMAGE_ATTRIBUTES = ['size', 'disk_format', 'owner',
                        'container_format', 'checksum', 'id',
                        'name', 'created_at', 'updated_at',
                        'deleted_at', 'deleted', 'status',
                        'min_disk', 'min_ram', 'is_public']
    output = {}
    for attr in IMAGE_ATTRIBUTES:
        output[attr] = getattr(image, attr, None)

    output['properties'] = getattr(image, 'properties', {})

    return output

def _remove_read_only(image_meta):
    IMAGE_ATTRIBUTES = ['status', 'updated_at', 'created_at', 'deleted_at']
    output = copy.deepcopy(image_meta)
    for attr in IMAGE_ATTRIBUTES:
        if attr in output:
            del output[attr]
    return output

def _reraise_translated_image_exception(image_id):
    """Transform the exception for the image but keep its traceback intact."""
    exc_type, exc_value, exc_trace = sys.exc_info()
    new_exc = _translate_image_exception(image_id, exc_value)
    raise new_exc, None, exc_trace

def _reraise_translated_exception():
    """Transform the exception but keep its traceback intact."""
    exc_type, exc_value, exc_trace = sys.exc_info()
    new_exc = _translate_plain_exception(exc_value)
    raise new_exc, None, exc_trace

def _translate_image_exception(image_id, exc_value):
    if isinstance(exc_value, (glanceclient.exc.Forbidden,
                              glanceclient.exc.Unauthorized)):
        return exception.ImageNotAuthorized(image_id=image_id)
    if isinstance(exc_value, glanceclient.exc.NotFound):
        return exception.ImageNotFound(image_id=image_id)
    if isinstance(exc_value, glanceclient.exc.BadRequest):
        return exception.Invalid(exc_value)
    return exc_value

def _translate_plain_exception(exc_value):
    if isinstance(exc_value, (glanceclient.exc.Forbidden,
                              glanceclient.exc.Unauthorized)):
        return exception.NotAuthorized(exc_value)
    if isinstance(exc_value, glanceclient.exc.NotFound):
        return exception.NotFound(exc_value)
    if isinstance(exc_value, glanceclient.exc.BadRequest):
        return exception.Invalid(exc_value)
    return exc_value

def get_remote_image_service(context, image_href):
    """Create an image_service and parse the id from the given image_href.

    The image_href param can be an href of the form
    'http://example.com:9292/v1/images/b8b2c6f7-7345-4e2f-afa2-eedaba9cbbe3',
    or just an id such as 'b8b2c6f7-7345-4e2f-afa2-eedaba9cbbe3'. If the
    image_href is a standalone id, then the default image service is returned.

    :param image_href: href that describes the location of an image
    :returns: a tuple of the form (image_service, image_id)

    """
    #NOTE(bcwaldon): If image_href doesn't look like a URI, assume its a
    # standalone image ID
    if '/' not in str(image_href):
        image_service = get_default_image_service()
        return image_service, image_href

    try:
        (image_id, glance_host, glance_port, use_ssl) = \
            _parse_image_ref(image_href)
        glance_client = GlanceClientWrapper(context=context,
                                            host=glance_host,
                                            port=glance_port,
                                            use_ssl=use_ssl)
    except ValueError:
        raise exception.InvalidImageRef(image_href=image_href)

    image_service = GlanceImageService(client=glance_client)
    return image_service, image_id

def get_default_image_service():
    return GlanceImageService()



# -*- coding: utf-8 -*-
# pep8-ignore: E501, E241

'''
====================================
 :mod:`ipcalc` IP subnet calculator
====================================
.. moduleauthor:: Wijnand Modderman-Lenstra <maze@pyth0n.org>
.. note:: BSD License

About
=====

This module allows you to perform network calculations.

References
==========

References:
 * http://www.estoile.com/links/ipv6.pdf
 * http://www.iana.org/assignments/ipv4-address-space
 * http://www.iana.org/assignments/multicast-addresses
 * http://www.iana.org/assignments/ipv6-address-space
 * http://www.iana.org/assignments/ipv6-tla-assignments
 * http://www.iana.org/assignments/ipv6-multicast-addresses
 * http://www.iana.org/assignments/ipv6-anycast-addresses

Thanks
======

Thanks to all who have contributed:

https://github.com/tehmaze/ipcalc/graphs/contributors
'''

__version__ = '1.1.2'

import warnings

try:
    bin
except NameError:
    def bin(x):
        '''
        Stringifies an int or long in base 2.
        '''
        if x < 0:
            return '-%s' % bin(-x)
        out = []
        if x == 0:
            out.append('0')
        while x > 0:
            out.append('01'[x & 1])
            x >>= 1
        out.reverse()
        return '0b%s' % ''.join(out)

class IP(object):
    '''
    Represents a single IP address.

    :param ip: the ip address
    :type ip: :class:`IP` or str or long or int

    >>> localhost = IP("127.0.0.1")
    >>> print localhost
    127.0.0.1
    >>> localhost6 = IP("::1")
    >>> print localhost6
    0000:0000:0000:0000:0000:0000:0000:0001
    '''

    # Hex-to-Bin conversion masks
    _bitmask = {
        '0': '0000', '1': '0001', '2': '0010', '3': '0011',
        '4': '0100', '5': '0101', '6': '0110', '7': '0111',
        '8': '1000', '9': '1001', 'a': '1010', 'b': '1011',
        'c': '1100', 'd': '1101', 'e': '1110', 'f': '1111'
    }

    # IP range specific information, see IANA allocations.
    _range = {
        4: {
            '01':                     'CLASS A',
            '10':                     'CLASS B',
            '110':                    'CLASS C',
            '1110':                   'CLASS D MULTICAST',
            '11100000':               'CLASS D LINKLOCAL',
            '1111':                   'CLASS E',
            '00001010':               'PRIVATE RFC1918',    # 10/8
            '101011000001':           'PRIVATE RFC1918',    # 172.16/12
            '1100000010101000':       'PRIVATE RFC1918',    # 192.168/16
        },
        6: {
            '00000000':               'RESERVED',           # ::/8
            '00000001':               'UNASSIGNED',         # 100::/8
            '0000001':                'NSAP',               # 200::/7
            '0000010':                'IPX',                # 400::/7
            '0000011':                'UNASSIGNED',         # 600::/7
            '00001':                  'UNASSIGNED',         # 800::/5
            '0001':                   'UNASSIGNED',         # 1000::/4
            '0010000000000000':       'RESERVED',           # 2000::/16 Reserved
            '0010000000000001':       'ASSIGNABLE',         # 2001::/16 Sub-TLA Assignments [RFC2450]
            '00100000000000010000000': 'ASSIGNABLE IANA',   # 2001:0000::/29 - 2001:01F8::/29 IANA
            '00100000000000010000001': 'ASSIGNABLE APNIC',  # 2001:0200::/29 - 2001:03F8::/29 APNIC
            '00100000000000010000010': 'ASSIGNABLE ARIN',   # 2001:0400::/29 - 2001:05F8::/29 ARIN
            '00100000000000010000011': 'ASSIGNABLE RIPE',   # 2001:0600::/29 - 2001:07F8::/29 RIPE NCC
            '0010000000000010':       '6TO4',               # 2002::/16 "6to4" [RFC3056]
            '0011111111111110':       '6BONE TEST',         # 3ffe::/16 6bone Testing [RFC2471]
            '0011111111111111':       'RESERVED',           # 3fff::/16 Reserved
            '010':                    'GLOBAL-UNICAST',     # 4000::/3
            '011':                    'UNASSIGNED',         # 6000::/3
            '100':                    'GEO-UNICAST',        # 8000::/3
            '101':                    'UNASSIGNED',         # a000::/3
            '110':                    'UNASSIGNED',         # c000::/3
            '1110':                   'UNASSIGNED',         # e000::/4
            '11110':                  'UNASSIGNED',         # f000::/5
            '111110':                 'UNASSIGNED',         # f800::/6
            '1111110':                'UNASSIGNED',         # fc00::/7
            '111111100':              'UNASSIGNED',         # fe00::/9
            '1111111010':             'LINKLOCAL',          # fe80::/10
            '1111111011':             'SITELOCAL',          # fec0::/10
            '11111111':               'MULTICAST',          # ff00::/8
            '0' * 96:                 'IPV4COMP',           # ::/96
            '0' * 80 + '1' * 16:      'IPV4MAP',            # ::ffff:0:0/96
            '0' * 128:                'UNSPECIFIED',        # ::/128
            '0' * 127 + '1':          'LOOPBACK'            # ::1/128
        }
    }

    def __init__(self, ip, mask=None, version=0):
        self.mask = mask
        self.v = 0
        # Parse input
        if ip is None:
            raise ValueError('Can not pass None')
        elif isinstance(ip, IP):
            self.ip = ip.ip
            self.dq = ip.dq
            self.v = ip.v
            self.mask = ip.mask
        elif isinstance(ip, (int, long)):
            self.ip = long(ip)
            if self.ip <= 0xffffffff:
                self.v = version or 4
                self.dq = self._itodq(ip)
            else:
                self.v = version or 6
                self.dq = self._itodq(ip)
        else:
            # If string is in CIDR or netmask notation
            if '/' in ip:
                ip, mask = ip.split('/', 1)
                self.mask = mask
            self.v = version or 0
            self.dq = ip
            self.ip = self._dqtoi(ip)
            assert self.v != 0, 'Could not parse input'
        # Netmask defaults to one ip
        if self.mask is None:
            self.mask = self.v == 4 and 32 or 128
        # Netmask is numeric CIDR subnet
        elif isinstance(self.mask, (int, long)) or self.mask.isdigit():
            self.mask = int(self.mask)
        # Netmask is in subnet notation
        elif isinstance(self.mask, basestring):
            limit = [32, 128][':' in self.mask]
            inverted = ~self._dqtoi(self.mask)
            count = 0
            while inverted & pow(2, count):
                count += 1
            self.mask = (limit - count)
        else:
            raise ValueError('Invalid netmask')
        # Validate subnet size
        if self.v == 6:
            self.dq = self._itodq(self.ip)
            if not 0 <= self.mask <= 128:
                raise ValueError('IPv6 subnet size must be between 0 and 128')
        elif self.v == 4:
            if not 0 <= self.mask <= 32:
                raise ValueError('IPv4 subnet size must be between 0 and 32')

    def bin(self):
        '''
        Full-length binary representation of the IP address.

        >>> ip = IP("127.0.0.1")
        >>> print ip.bin()
        01111111000000000000000000000001
        '''
        return bin(self.ip).split('b')[1].rjust(self.mask, '0')

    def hex(self):
        '''
        Full-length hexadecimal representation of the IP address.

        >>> ip = IP("127.0.0.1")
        >>> print ip.hex()
        7f000001
        '''
        if self.v == 4:
            return '%08x' % self.ip
        else:
            return '%032x' % self.ip

    def subnet(self):
        return self.mask

    def version(self):
        '''
        IP version.

        >>> ip = IP("127.0.0.1")
        >>> print ip.version()
        4
        '''
        return self.v

    def info(self):
        '''
        Show IANA allocation information for the current IP address.

        >>> ip = IP("127.0.0.1")
        >>> print ip.info()
        CLASS A
        '''
        b = self.bin()
        self.v == 4 and 32 or 128
        for i in range(len(b), 0, -1):
            if b[:i] in self._range[self.v]:
                return self._range[self.v][b[:i]]
        return 'UNKNOWN'

    def _dqtoi(self, dq):
        '''
        Convert dotquad or hextet to long.
        '''
        # hex notation
        if dq.startswith('0x'):
            ip = long(dq[2:], 16)
            if ip > 0xffffffffffffffffffffffffffffffffL:
                raise ValueError('%s: IP address is bigger than 2^128' % dq)
            if ip <= 0xffffffff:
                self.v = 4
            else:
                self.v = 6
            return ip

        # IPv6
        if ':' in dq:
            # Split hextets
            hx = dq.split(':')
            if ':::' in dq:
                raise ValueError("%s: IPv6 address can't contain :::" % dq)
            # Mixed address (or 4-in-6), ::ffff:192.0.2.42
            if '.' in dq:
                return self._dqtoi(hx[-1])
            if len(hx) > 8:
                raise ValueError('%s: IPv6 address with more than 8 hexlets' % dq)
            elif len(hx) < 8:
                # No :: in address
                if not '' in hx:
                    raise ValueError('%s: IPv6 address invalid: '
                                     'compressed format malformed' % dq)
                elif not (dq.startswith('::') or dq.endswith('::')) and len([x for x in hx if x == '']) > 1:
                    raise ValueError('%s: IPv6 address invalid: '
                                     'compressed format malformed' % dq)
                ix = hx.index('')
                px = len(hx[ix + 1:])
                for x in xrange(ix + px + 1, 8):
                    hx.insert(ix, '0')
            elif dq.endswith('::'):
                pass
            elif '' in hx:
                raise ValueError('%s: IPv6 address invalid: '
                                 'compressed format detected in full notation' % dq())
            ip = ''
            hx = [x == '' and '0' or x for x in hx]
            for h in hx:
                if len(h) < 4:
                    h = '%04x' % int(h, 16)
                if not 0 <= int(h, 16) <= 0xffff:
                    raise ValueError('%r: IPv6 address invalid: '
                                     'hexlets should be between 0x0000 and 0xffff' % dq)
                ip += h
            self.v = 6
            return long(ip, 16)
        elif len(dq) == 32:
            # Assume full heximal notation
            self.v = 6
            return long(h, 16)

        # IPv4
        if '.' in dq:
            q = dq.split('.')
            q.reverse()
            if len(q) > 4:
                raise ValueError('%s: IPv4 address invalid: '
                                 'more than 4 bytes' % dq)
            for x in q:
                if not 0 <= int(x) <= 255:
                    raise ValueError('%s: IPv4 address invalid: '
                                     'bytes should be between 0 and 255' % dq)
            while len(q) < 4:
                q.insert(1, '0')
            self.v = 4
            return sum(long(byte) << 8 * index for index, byte in enumerate(q))

        raise ValueError('Invalid address input')

    def _itodq(self, n):
        '''
        Convert long to dotquad or hextet.
        '''
        if self.v == 4:
            return '.'.join(map(str, [
                (n >> 24) & 0xff,
                (n >> 16) & 0xff,
                (n >> 8) & 0xff,
                n & 0xff,
            ]))
        else:
            n = '%032x' % n
            return ':'.join(n[4 * x:4 * x + 4] for x in xrange(0, 8))

    def __str__(self):
        '''
        Return dotquad representation of the IP.

        >>> ip = IP("::1")
        >>> print str(ip)
        0000:0000:0000:0000:0000:0000:0000:0001
        '''
        return self.dq

    def __repr__(self):
        '''
        Return canonical representation of the IP.

        >>> ip = IP("::1")
        >>> print repr(ip)
        IP('0000:0000:0000:0000:0000:0000:0000:0001', mask=128, version=6)
        '''
        return "%s('%s', mask=%d, version=%d)" % (
                        self.__class__.__name__,
                        self.dq,
                        self.mask,
                        self.v)

    def __hash__(self):
        return hash(self.to_tuple())

    hash = __hash__

    def __int__(self):
        return int(self.ip)

    def __long__(self):
        return self.ip

    def __lt__(self, other):
        return long(self) < long(IP(other))

    def __le__(self, other):
        return long(self) <= long(IP(other))

    def __ge__(self, other):
        return long(self) >= long(IP(other))

    def __gt__(self, other):
        return long(self) > long(IP(other))

    def __eq__(self, other):
        return long(self) == long(IP(other))

    def size(self):
        return 1

    def clone(self):
        '''
        Return a new <IP> object with a copy of this one.

        >>> ip = IP('127.0.0.1')
        >>> ip2 = ip.clone()
        >>> ip2
        IP('127.0.0.1', mask=32, version=4)
        >>> ip is ip2
        False
        >>> ip == ip2
        True
        >>> ip.mask = 24
        >>> ip2.mask
        32
        '''
        return IP(self)

    def to_ipv4(self):
        '''
        Convert (an IPv6) IP address to an IPv4 address, if possible. Only works
        for IPv4-compat (::/96), IPv4-mapped (::ffff/96), and 6-to-4 (2002::/16) addresses.

        >>> ip = IP('2002:c000:022a::')
        >>> print ip.to_ipv4()
        192.0.2.42
        '''
        if self.v == 4:
            return self
        else:
            if self.bin().startswith('0' * 96):
                return IP(long(self), version=4)
            elif self.bin().startswith('0' * 80 + '1' * 16):
                return IP(long(self) & 0xffffffff, version=4)
            elif long(self) & 0x20020000000000000000000000000000L:
                return IP((long(self) - 0x20020000000000000000000000000000L) >> 80, version=4)
            else:
                return ValueError('%s: IPv6 address is not IPv4 compatible or mapped, '
                                  'nor an 6-to-4 IP' % self.dq)

    @classmethod
    def from_bin(cls, value):
        value = value.lstrip('b')
        if len(value) == 32:
            return cls(int(value, 2))
        elif len(value) == 128:
            return cls(long(value, 2))
        else:
            return ValueError('%r: invalid binary notation' % (value,))

    @classmethod
    def from_hex(cls, value):
        if len(value) == 8:
            return cls(int(value, 16))
        elif len(value) == 32:
            return cls(long(value, 16))
        else:
            raise ValueError('%r: invalid hexadecimal notation' % (value,))

    def to_ipv6(self, type='6-to-4'):
        '''
        Convert (an IPv4) IP address to an IPv6 address.

        >>> ip = IP('192.0.2.42')
        >>> print ip.to_ipv6()
        2002:c000:022a:0000:0000:0000:0000:0000

        >>> print ip.to_ipv6('compat')
        0000:0000:0000:0000:0000:0000:c000:022a

        >>> print ip.to_ipv6('mapped')
        0000:0000:0000:0000:0000:ffff:c000:022a
        '''
        assert type in ['6-to-4', 'compat', 'mapped'], 'Conversion type not supported'
        if self.v == 4:
            if type == '6-to-4':
                return IP(0x20020000000000000000000000000000L | long(self) << 80, version=6)
            elif type == 'compat':
                return IP(long(self), version=6)
            elif type == 'mapped':
                return IP(0xffff << 32 | long(self), version=6)
        else:
            return self

    def to_reverse(self):
        '''
        Convert the IP address to a PTR record in .in-addr.arpa for IPv4 and
        .ip6.arpa for IPv6 addresses.

        >>> ip = IP('192.0.2.42')
        >>> print ip.to_reverse()
        42.2.0.192.in-addr.arpa
        >>> print ip.to_ipv6().to_reverse()
        0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.a.2.2.0.0.0.0.c.2.0.0.2.ip6.arpa
        '''
        if self.v == 4:
            return '.'.join(list(self.dq.split('.')[::-1]) + ['in-addr', 'arpa'])
        else:
            return '.'.join(list(self.hex())[::-1] + ['ip6', 'arpa'])

    def to_tuple(self):
        '''
        Used for comparisons.
        '''
        return (self.dq, self.mask)

class Network(IP):
    '''
    Network slice calculations.

    :param ip: network address
    :type ip: :class:`IP` or str or long or int
    :param mask: netmask
    :type mask: int or str

    >>> localnet = Network('127.0.0.1/8')
    >>> print localnet
    127.0.0.1
    '''

    def netmask(self):
        '''
        Network netmask derived from subnet size, as IP object.

        >>> localnet = Network('127.0.0.1/8')
        >>> print localnet.netmask()
        255.0.0.0
        '''
        return IP(self.netmask_long(), version=self.version())

    def netmask_long(self):
        '''
        Network netmask derived from subnet size, as long.

        >>> localnet = Network('127.0.0.1/8')
        >>> print localnet.netmask_long()
        4278190080
        '''
        if self.version() == 4:
            return (0xffffffffL >> (32 - self.mask)) << (32 - self.mask)
        else:
            return (0xffffffffffffffffffffffffffffffffL >> (128 - self.mask)) << (128 - self.mask)

    def network(self):
        '''
        Network address, as IP object.

        >>> localnet = Network('127.128.99.3/8')
        >>> print localnet.network()
        127.0.0.0
        '''
        return IP(self.network_long(), version=self.version())

    def network_long(self):
        '''
        Network address, as long.

        >>> localnet = Network('127.128.99.3/8')
        >>> print localnet.network_long()
        2130706432
        '''
        return self.ip & self.netmask_long()

    def broadcast(self):
        '''
        Broadcast address, as IP object.

        >>> localnet = Network('127.0.0.1/8')
        >>> print localnet.broadcast()
        127.255.255.255
        '''
        # XXX: IPv6 doesn't have a broadcast address, but it's used for other
        #      calculations such as <Network.host_last>
        return IP(self.broadcast_long(), version=self.version())

    def broadcast_long(self):
        '''
        Broadcast address, as long.

        >>> localnet = Network('127.0.0.1/8')
        >>> print localnet.broadcast_long()
        2147483647
        '''
        if self.version() == 4:
            return self.network_long() | (0xffffffffL - self.netmask_long())
        else:
            return self.network_long() \
                | (0xffffffffffffffffffffffffffffffffL - self.netmask_long())

    def host_first(self):
        '''
        First available host in this subnet.
        '''
        if (self.version() == 4 and self.mask > 30) or \
                (self.version() == 6 and self.mask > 126):
            return self
        else:
            return IP(self.network_long() + 1, version=self.version())

    def host_last(self):
        '''
        Last available host in this subnet.
        '''
        if (self.version() == 4 and self.mask == 32) or \
                (self.version() == 6 and self.mask == 128):
            return self
        elif (self.version() == 4 and self.mask == 31) or \
                (self.version() == 6 and self.mask == 127):
            return IP(long(self) + 1, version=self.version())
        else:
            return IP(self.broadcast_long() - 1, version=self.version())

    def in_network(self, other):
        '''
        Check if the given IP address is within this network.
        This function is deprecated, use check_collision instead.
        '''
        warnings.warn('%s.in_network is deprecated, use check_collision in '
                      'stead' % (self.__class__.__name__,),
                      DeprecationWarning,
                      stacklevel=2)
        other = Network(other)
        return self.network_long() <= other.network_long() <= self.broadcast_long()

    def check_collision(self, other):
        '''
        Check another network against the given network and checks for IP collisions.
        '''
        other = Network(other)
        return self.network_long() <= other.network_long() <= self.broadcast_long() or \
            other.network_long() <= self.network_long() <= other.broadcast_long()

    def __contains__(self, ip):
        '''
        Check if the given ip is part of the network.

        >>> '192.0.2.42' in Network('192.0.2.0/24')
        True
        >>> '192.168.2.42' in Network('192.0.2.0/24')
        False
        '''
        return self.check_collision(ip)

    def __lt__(self, other):
        return self.size() < IP(other).size()

    def __le__(self, other):
        return self.size() <= IP(other).size()

    def __gt__(self, other):
        return self.size() > IP(other).size()

    def __ge__(self, other):
        return self.size() >= IP(other).size()

    def __eq__(self, other):
        return self.size() == IP(other).size()

    def __getitem__(self, key):
        if isinstance(key, slice):
            # Work-around IPv6 subnets being huge. Slice indices don't like
            # long int.
            x = key.start or 0
            slice_stop = (key.stop or self.size()) - 1
            slice_step = key.step or 1
            arr = list()
            while x < slice_stop:
                arr.append(IP(long(self) + x))
                x += slice_step
            return tuple(arr)
        else:
            return IP(long(self) + key)

    def __iter__(self):
        '''
        Generate a range of usable host IP addresses within the network, as IP
        objects.

        >>> for ip in Network('192.168.114.0/30'):
        ...     print str(ip)
        ...
        192.168.114.1
        192.168.114.2
        '''

        curr = long(self.host_first())
        stop = long(self.host_last())
        while curr <= stop:
            yield IP(curr)
            curr += 1

    def has_key(self, ip):
        '''
        Check if the given ip is part of the network.

        :param ip: the ip address
        :type ip: :class:`IP` or str or long or int

        >>> net = Network('192.0.2.0/24')
        >>> net.has_key('192.168.2.0')
        False
        >>> net.has_key('192.0.2.42')
        True
        '''
        return self.__contains__(ip)

    def size(self):
        '''
        Number of ip's within the network.

        >>> net = Network('192.0.2.0/24')
        >>> print net.size()
        256
        '''
        return 2 ** ((self.version() == 4 and 32 or 128) - self.mask)

if __name__ == '__main__':
    tests = [
        ('192.168.114.42', 23, ['192.168.0.1', '192.168.114.128', '10.0.0.1']),
        ('123::', 128, ['123:456::', '::1', '123::456']),
        ('::42', 64, ['::1', '1::']),
        ('2001:dead:beef:1:c01d:c01a::', 48, ['2001:dead:beef:babe::']),
        ('10.10.0.0', '255.255.255.0', ['10.10.0.20', '10.10.10.20']),
        ('2001:dead:beef:1:c01d:c01a::', 'ffff:ffff:ffff::', ['2001:dead:beef:babe::']),
        ('10.10.0.0/255.255.240.0', None, ['10.10.0.20', '10.10.250.0']),
    ]

    for ip, mask, test_ip in tests:
        net = Network(ip, mask)
        print '==========='
        print 'ip address:', net
        print 'to ipv6...:', net.to_ipv6()
        print 'ip version:', net.version()
        print 'ip info...:', net.info()
        print 'subnet....:', net.subnet()
        print 'num ip\'s..:', net.size()
        print 'integer...:', long(net)
        print 'hex.......:', net.hex()
        print 'netmask...:', net.netmask()
        # Not implemented in IPv6
        if net.version() == 4:
            print 'network...:', net.network()
            print 'broadcast.:', net.broadcast()
        print 'first host:', net.host_first()
        print 'reverse...:', net.host_first().to_reverse()
        print 'last host.:', net.host_last()
        print 'reverse...:', net.host_last().to_reverse()
        for ip in test_ip:
            print '%s in network: ' % ip, ip in net



# Copyright (c) 2011-2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Filter support
"""

import inspect

from stevedore import extension

class BaseFilter(object):
    """Base class for all filter classes."""
    def _filter_one(self, obj, filter_properties):
        """Return True if it passes the filter, False otherwise.
        Override this in a subclass.
        """
        return True

    def filter_all(self, filter_obj_list, filter_properties):
        """Yield objects that pass the filter.

        Can be overriden in a subclass, if you need to base filtering
        decisions on all objects.  Otherwise, one can just override
        _filter_one() to filter a single object.
        """
        for obj in filter_obj_list:
            if self._filter_one(obj, filter_properties):
                yield obj

class BaseFilterHandler(object):
    """ Base class to handle loading filter classes.

    This class should be subclassed where one needs to use filters.
    """
    def __init__(self, filter_class_type, filter_namespace):
        self.namespace = filter_namespace
        self.filter_class_type = filter_class_type
        self.filter_manager = extension.ExtensionManager(filter_namespace)

    def _is_correct_class(self, obj):
        """Return whether an object is a class of the correct type and
        is not prefixed with an underscore.
        """
        return (inspect.isclass(obj) and
                not obj.__name__.startswith('_') and
                issubclass(obj, self.filter_class_type))

    def get_all_classes(self):
        return [x.plugin for x in self.filter_manager
                if self._is_correct_class(x.plugin)]

    def get_filtered_objects(self, filter_classes, objs,
                             filter_properties):
        for filter_cls in filter_classes:
            objs = filter_cls().filter_all(objs, filter_properties)
        return list(objs)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 Intel.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# pylint: disable=W0621

"""
Conductor Service
"""

import json
from oslo.config import cfg
import datetime
from vsm import context
from vsm import db
from vsm import exception
from vsm import flags
from vsm import manager
from vsm.openstack.common import excutils
from vsm.openstack.common import importutils
from vsm.openstack.common import log as logging
from vsm.openstack.common.notifier import api as notifier
from vsm.openstack.common import timeutils

LOG = logging.getLogger(__name__)
FLAGS = flags.FLAGS

class ConductorManager(manager.Manager):
    """Chooses a host to create storages."""

    RPC_API_VERSION = '1.2'

    def __init__(self, service_name=None, *args, **kwargs):
        #if not scheduler_driver:
        #    scheduler_driver = FLAGS.scheduler_driver
        #self.driver = importutils.import_object(scheduler_driver)
        super(ConductorManager, self).__init__(*args, **kwargs)

    def init_host(self):
        LOG.info('init_host in manager ')

    def test_service(self, context):
        LOG.info(' test_service in conductor')
        return {'key': 'test_server_in_conductor'}

    def check_poolname(self, context, poolname):
        pool_list = db.pool_get_all(context)
        if pool_list:
            for pool in pool_list:
                if pool['name'] == poolname:
                    return 1
        return 0

    def create_storage_pool(self, context, body):
        #TO BE DONE
        body['cluster_id'] = db.cluster_get_all(context)[0]['id']#1
        res = db.pool_create(context, body)
        return res

    def update_storage_pool(self, context, pool_id, values):
        return db.pool_update(context, pool_id, values)

    def update_storage_pool_by_name(self, context, pool_name, cluster_id, values):
        return db.pool_update_by_name(context, pool_name, cluster_id, values)

    def get_osd_num(self, context, group_id):
        osds = db.osd_get_all(context)

        osd_num = 0
        for osd in osds:
            if osd['deleted'] == False and \
            str(osd['storage_group_id']) == str(group_id):
                osd_num = osd_num + 1

        return osd_num

    def list_storage_pool(self, context):
        LOG.info('list_storage_pool in conductor manager')

        pool_list = db.pool_get_all(context)
        pool_list_dict = {}
        if pool_list:
            for pool in pool_list:
                pool_list_dict[pool['id']] = pool

        return pool_list_dict

    def get_storage_pool(self, context, id):
        LOG.info('get_storage_pool in conductor manager')
        return db.pool_get_by_db_id(context, id)

    def destroy_storage_pool(self, context, pool_name):
        if pool_name:
            db.pool_destroy(context, pool_name)

    def get_storage_group_list(self, context):
        LOG.info('get_storage_group_list in conductor manager')

        storage_group_list = db.storage_group_get_all(context)
        storage_group_list = [x for x in storage_group_list if x.status == "IN"]
        # osds = db.osd_get_all(context)

        group_list = {}
        group_name_list = []
        if storage_group_list:
            for group in storage_group_list:
                # osd_num=0
                # for osd in osds:
                #     if osd['storage_group_id'] == group['id']:
                #         osd_num = osd_num + 1
                # if osd_num > 0:
                if group['name'] not in group_name_list:
                    group_name_list.append(group['name'])
                    group_list[group['id']] = group['name']
        return group_list

    def get_server_list(self, context):
        LOG.info('get_server_list in conductor manager')
        server_list = db.init_node_get_all(context)
        ret = self._set_error(context)
        for ser in server_list:
            if ret:
                ser['status'] = ret
        return server_list

    def _set_error(self, context, cluster_id = None):
        if cluster_id:
            summary = db.summary_get_by_cluster_id_and_type(context, cluster_id, 'cluster')
        else:
            summary = db.summary_get_by_type_first(context, 'cluster')
        #summary = db.summary_get_by_cluster_id_and_type(context, 1, 'cluster')
        if summary:
            sum_data = json.loads(summary['summary_data'])
            h_list = sum_data.get('health_list')
            if len(h_list) > 0:
                if h_list[0].find('ERROR') != -1:
                    return h_list[0]
        return None

    def ceph_error(self, context):
        return self._set_error(context)

    def get_cluster_list(self, context):
        LOG.info('get_cluster_list in conductor manager')
        cluster_list = db.init_node_get_all(context)
        ret = self._set_error(context)
        for ser in cluster_list:
            if ret:
                ser['status'] = 'unavailable'
        return cluster_list

    def get_server(self, context, id):
        LOG.info('get_server in conductor manager')
        server = db.init_node_get(context, id)
        ret = self._set_error(context)
        if ret:
            server['status'] = 'unavailable'

        LOG.info("CEPH_LOG log server %s" % server)
        return server

    def get_zone_list(self, context):
        LOG.info('get_zone_list in conductor manager')
        zone_list = db.zone_get_all(context)
        LOG.info("CEPH_LOG log server_list %s" % zone_list)
        return zone_list

    def get_mapping(self, context):
        LOG.info('get_mapping in conductor manager')
        storage_group_list = db.storage_group_get_all(context)
        mapping = {}
        if storage_group_list:
            for group in storage_group_list:
                mapping[group['rule_id']] = group['friendly_name']
        return mapping

    def get_ruleset_id(self, context, group_id):
        LOG.info("Get ruleset id via storage_group id.")
        storage_group = db.storage_group_get(context, group_id)
        return storage_group['rule_id']

    def count_hosts_by_storage_group_id(self, context, storage_group_id):
        return db.osd_state_count_service_id_by_storage_group_id(context, \
           storage_group_id)

    def init_node_get_by_host(self, context, host):
        """Get init node by host name."""
        return db.init_node_get_by_host(context, host)

    def init_node_get_by_cluster_id(self, context, cluster_id):
        """Get init node by cluster id."""
        return db.init_node_get_by_cluster_id(context, cluster_id)

    def init_node_get_cluster_nodes(self, context, init_node_id):
        """Get cluster nodes by id"""
        return db.init_node_get_cluster_nodes(context, init_node_id)

    #init_node
    def init_node_get_by_id_and_type(self, context, id, type):
        #LOG.info("get init node by id and type")
        init_node = db.init_node_get_by_id_and_type(context, id, type)
        return init_node

    def init_node_get_by_id(self, context, id):
        #LOG.info("get init node by id")
        init_node = db.init_node_get_by_id(context, id)
        return init_node

    def init_node_create(self, context, values):
        return db.init_node_create(context, values)

    def init_node_update(self, context, id, values):
        return db.init_node_update(context, id, values)

    def init_node_get_by_primary_public_ip(self, context, primary_public_ip):
        return db.init_node_get_by_primary_public_ip(context, \
                                                     primary_public_ip)

    def init_node_get_by_secondary_public_ip(self, context, \
                                             secondary_public_ip):
        return db.init_node_get_by_secondary_public_ip(context, \
                                                       secondary_public_ip)

    def init_node_get_by_cluster_ip(self, context, cluster_ip):
        return db.init_node_get_by_cluster_ip(context, cluster_ip)

    def init_node_update_status_by_id(self,
                                      context,
                                      init_node_id,
                                      status):
        """ConductorManager update the status of init node."""
        return db.init_node_update_status_by_id(context,
                                                init_node_id,
                                                status)

    #osd_state
    def osd_get(self, context, osd_id):
        return db.osd_get(context, osd_id)

    def osd_delete(self, context, osd_id):
        return db.osd_delete(context, osd_id)

    def osd_remove(self, context, osd_id):
        return db.osd_remove(context, osd_id)

    def osd_state_get_all(self,
                          context,
                          limit=None,
                          marker=None,
                          sort_keys=None,
                          sort_dir=None,
                          search_opts={}):
        all_osd = db.osd_state_get_all(context,
                                       limit,
                                       marker,
                                       sort_keys,
                                       sort_dir,
                                       search_opts)
        return all_osd

    def osd_state_get_by_name(self, context, name):
        return db.osd_state_get_by_name(context, name)

    def osd_state_create(self, context, values):
        LOG.info('ADD_OSD values = %s' % values)
        result = db.osd_state_get_by_osd_name_and_service_id_and_cluster_id(\
                 context, values['osd_name'], values['service_id'],\
                 values['cluster_id'])
        if not result:
            result = db.osd_state_get_by_device_id_and_service_id_and_cluster_id(\
                 context, values['device_id'], values['service_id'],\
                 values['cluster_id'])
        LOG.info('ADD_OSD result = %s' % result)

        if not result:
            LOG.info('ADD_OSD result is None')
            return db.\
               osd_state_create(context, values)
        else:
            LOG.info('ADD_OSD result is ok')
            values['id'] = result['id']
            values['deleted'] = 0
            return db.\
                osd_state_update(context, values['id'], values)

    def osd_state_update(self, context, values):
        #LOG.info("osd_state_update")
        osd_ref = db.osd_state_get_by_name(context, values['osd_name'])
        if osd_ref:
            osd_state = db.osd_state_update(context, osd_ref['id'], values)
            return osd_state

    def osd_state_update_or_create(self, context, values, create=None):
        #LOG.info("osd_state_update_or_create---%s"%values)
        osd_ref = db.osd_state_get_by_name(context, values['osd_name'])
        if not osd_ref:
            osd_ref = db.osd_state_get_by_device_id_and_service_id_and_cluster_id(\
                 context, values['device_id'], values['service_id'],\
                 values['cluster_id'])
        if osd_ref:
            osd_state = db.osd_state_update(context, osd_ref['id'], values)
            return osd_state
        else:
            create = True

        if create is None:
            osd_state = db.osd_state_update_or_create(context, values)
        elif create == True:
            osd_state = db.osd_state_create(context, values)
        else:
            #LOG.info("osd values:%s" % values)
            osd_state_ref = db.osd_state_get_by_name(context, values['osd_name'])
            if osd_state_ref:
                values['id'] = osd_state_ref.id
                osd_state = db.osd_state_update(context,
			                      values['id'],
					      values)
            else:
                return None 
        return osd_state

    def osd_state_count_by_init_node_id(self, context, init_node_id):
        return db.osd_state_count_by_init_node_id(context, init_node_id)

    def osd_state_get_by_service_id_and_storage_group_id(self, context, \
                                                         service_id, \
                                                         storage_group_id):
        return db.osd_state_get_by_service_id_and_storage_group_id(\
                  context, service_id, storage_group_id)

    def osd_state_get_by_service_id(self, context, service_id):
        return db.osd_state_get_by_service_id(context, service_id)

    def osd_state_get_by_osd_name_and_service_id_and_cluster_id(self, \
            context, osd_name, service_id, cluster_id):
        return db.osd_state_get_by_osd_name_and_service_id_and_cluster_id(\
            context, osd_name, service_id, cluster_id)

    def osd_state_get_by_device_id_and_service_id_and_cluster_id(self, \
            context, device_id, service_id, cluster_id):
        return db.osd_state_get_by_device_id_and_service_id_and_cluster_id(\
            context, device_id, service_id, cluster_id)
    #device
    def device_get_all(self, context):
        error = self._set_error(context)
        all_devices = db.device_get_all(context)
        if not error:
            return all_devices

        for dev in all_devices:
            dev['state'] = error
            dev['journal_state'] = error

        return all_devices

    def device_get_by_hostname(self, context, hostname):
        init_node = db.init_node_get_by_hostname(context, hostname)
        if init_node:
            service_id = init_node['service_id']
            device_list = db.device_get_by_service_id(context, service_id)
            if device_list:
                return device_list
        else:
            return None

    def device_create(self, context, values):
        return db.device_create(context, values)

    def device_update_or_create(self, context, values, create=None):
        if create is None:
            device = db.device_update_or_create(context, values)
        elif create is True:
            device = db.device_create(context, values)
        else:
            device = db.device_update(context,\
                                      values['id'],\
                                      values)
        return device

    def device_get_all_by_service_id(self, context, service_id):
        return db.device_get_all_by_service_id(context, service_id)

    def device_get_distinct_storage_class_by_service_id(self, context,\
                                                        service_id):
        return db.device_get_distinct_storage_class_by_service_id(context,\
                                                                 service_id)

    def device_get_by_name_and_journal_and_service_id(self, context, \
                                                  name, journal, service_id):
        return db.device_get_by_name_and_journal_and_service_id(context, name, \
                                                journal, service_id)

    #storage_group
    def storage_group_get_all(self, context):
        return db.storage_group_get_all(context)

    def storage_group_get(self, context, id):
        return db.storage_group_get(context, id)

    def create_storage_group(self, context, values):
        if values is None:
            LOG.warn("Error: Empty values")
            try:
                raise exception.GetNoneError
            except exception.GetNoneError, e:
                LOG.error("%s:%s", e.code, e.message)
            return False

        res = db.storage_group_get_all(context)
        name_list = []
        for item in res:
            name_list.append(item['name'])

        if values['name'] not in name_list:
            db.storage_group_create(context, values)
        else:
            LOG.info('Warnning: name exists in table %s' % values['name'])
            return False

        return True

    #zone
    def create_zone(self, context, values=None):
        if values is None:
            LOG.warn("Error: Empty values")
            try:
                raise exception.GetNoneError
            except exception.GetNoneError, e:
                LOG.error("%s:%s", e.code, e.message)
            return False

        res = db.zone_get_all(context)
        zone_list = []
        for item in res:
            zone_list.append(item['name'])

        if values['name'] not in zone_list:
            db.zone_create(context, values)
        else:
            LOG.info('Warnning: zone exists in table %s' % values['name'])
            return True

        return True

    def zone_get_all(self, context):
        return db.zone_get_all(context)

    def zone_get_by_id(self, context, id):
        return db.zone_get_by_id(context, id)

    def zone_get_by_name(self, context, name):
        return db.zone_get_by_name(context, name)

    #cluster
    def cluster_create(self, context, values):
        return db.cluster_create(context, values)

    def cluster_update(self, context, cluster_id, values):
        return db.cluster_update(context, cluster_id, values)

    def cluster_get_by_name(self, context, name):
        return db.cluster_get_by_name(context, name)

    def cluster_get_all(self, context):
        return db.cluster_get_all(context)

    def cluster_info_dict_get_by_id(self, context, cluster_id):
        return db.cluster_info_dict_get_by_id(context, cluster_id)

    #service
    def service_get_by_host_and_topic(self, context, host, topic):
        return db.service_get_by_host_and_topic(context, host, topic)

    #ceph
    def host_devices_by_init_node_id(self,
                                     context,
                                     init_node_id):
        init_node = db.init_node_get_by_id(context, init_node_id)
        #LOG.info("host_devices:init_node:%s", init_node)
        zone = db.zone_get_by_id(context, init_node['zone_id'])
        devices = db.device_get_all_by_service_id(context,
                                                  init_node['service_id'])
        file_system = init_node.cluster.file_system
        #osd_heartbeat_interval=init_node.cluster.osd_heartbeat_interval
        #osd_heartbeat_grace=init_node.cluster.osd_heartbeat_grace

        lst = []
        for device in devices:
            d = {}
            d['host'] = init_node['host']
            d['cluster_id'] = init_node['cluster_id']
            d['file_system'] = file_system
            d['primary_public_ip'] = init_node['primary_public_ip']
            d['secondary_public_ip'] = init_node['secondary_public_ip']
            d['cluster_ip'] = init_node['cluster_ip']
            d['data_drives_number'] = init_node['data_drives_number']
            d['dev_name'] = device['name']
            d['dev_journal'] = device['journal']
            d['dev_id'] = device['id']
            d['service_id'] = device['service_id']
            d['storage_class'] = device['device_type']
            d['zone_id'] = init_node['zone_id']
            d['zone'] = zone['name']
            lst.append(d)

        return lst

    def host_storage_groups_devices(self,
                                    context,
                                    init_node_id):
        # TODO all code refered is used as type = 'storage'
        # What type means?
        # Need to put the implementation into DB or manager.
        LOG.info(' conductor api:host_storage_groups_devices()')
        lst = self.host_devices_by_init_node_id(context, \
               init_node_id)

        storage_group_list = db.storage_group_get_all(context)

        for item in lst:
            for storage_group in storage_group_list:
                if storage_group['storage_class'] == item['storage_class']:
                    item['storage_group_id'] = storage_group['id']
                    # TODO should we use storage_group_name instead?
                    item['storage_group'] = storage_group['name']
                    item['storage_group_name'] = storage_group['name']
        return lst

    def ceph_node_info(self, context, init_node_id):
        init_node = db.init_node_get_by_id(context, init_node_id)
        zone = db.zone_get_by_id(context, init_node['zone_id'])
        storage_class_list = db.\
                             device_get_distinct_storage_class_by_service_id(\
                             context, init_node['service_id'])
        #LOG.info('storage_class_list:%s', storage_class_list)
        storage_group_list = db.storage_group_get_all(context)
        #LOG.info('storage_group_list:%s', storage_group_list)
        list = []
        for storage_class in storage_class_list:
            dict = {}
            for storage_group in storage_group_list:
                if storage_class == storage_group['storage_class']:
                    dict['storage_group_id'] = storage_group['id']
                    dict['storage_group_name'] = storage_group['name']
                    break
            list.append(dict)
        #LOG.info('list:%s', list)
        final_list = []
        for item in list:
            osd_state_list = db.\
                             osd_state_get_by_service_id_and_storage_group_id(  \
                             context, init_node['service_id'], \
                             item['storage_group_id'])
            if not osd_state_list:
                continue
            #LOG.info('osd_state_list:%s', osd_state_list)
            for osd_state in osd_state_list:
                final_dict = {}
                final_dict['osd_state_id'] = osd_state['id']
                final_dict['osd_state_name'] = osd_state['osd_name']
                final_dict['storage_group_name'] = item['storage_group_name']
                final_dict['host'] = init_node['host']
                final_dict['zone'] = zone['name']
                final_list.append(final_dict)
        return final_list

    #pg
    def pg_get_all(self, context, limit, marker, sort_keys, sort_dir):
        return db.pg_get_all(context, limit, marker, sort_keys, sort_dir)

    def pg_get(self, context, id):
        return db.pg_get(context, id)

    def pg_update_or_create(self, context, values):
        pg_ref = db.pg_get_by_pgid(context, values['pgid'])
        if pg_ref:
            pg = db.pg_update(context, pg_ref['id'], values)
        else:
            pg = db.pg_create(context, values)
        return pg

    #rbd
    def rbd_get_all(self, context, limit, marker, sort_keys, sort_dir):
        return db.rbd_get_all(context, limit, marker, sort_keys, sort_dir)

    def rbd_get(self, context, id):
        return db.rbd_get(context, id)

    def license_status_create(self, context, values):
        return db.license_status_create(context, values)

    def license_status_get(self, context):
        return db.license_status_get(context)

    def license_status_update(self, context, value):
        return db.license_status_update(context, value)

    #mds
    def mds_get_all(self, context):
        return db.mds_get_all(context)

    def mds_get(self, context, id):
        return db.mds_get_by_id(context, id)
    
    def zones_hosts_get_by_storage_group(self, context, storage_group):
        osds = db.osd_state_get_all(context)
        host_list = []
        zone_list = []
        for osd in osds:
            if storage_group == osd['storage_group']['name']:
                if osd['zone']['name'] not in zone_list:
                    zone_list.append(osd['zone']['name'])
                if osd['service']['host'] not in host_list:
                    host_list.append(osd['service']['host']) 
        return zone_list, host_list

    def get_performance_metrics(self, context, search_opts):
        return db.get_performance_metrics(context, search_opts=search_opts)

    def get_sum_performance_metrics(self, context, search_opts):
        return db.get_sum_performance_metrics(context, search_opts=search_opts)

    def get_latency(self, context, search_opts):
        return db.get_latency(context, search_opts=search_opts)

    def get_cpu_usage(self, context, search_opts):
        return db.get_cpu_usage(context, search_opts=search_opts)

    def get_poolusage(self, context, poolusage_id):
        return db.get_poolusage(context, poolusage_id=poolusage_id)

    def get_appnode(self, context, appnode_id):
        return db.appnodes_get_by_id(context, appnode_id)

    def delete_pool_usage(self, context, poolusage_id):
        return db.destroy_storage_pool_usage(context, poolusage_id)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2010 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from lxml import etree

from vsm.api.openstack import wsgi
from vsm.api.views import versions as views_versions
from vsm.api import xmlutil

VERSIONS = {
    "v2.0": {
        "id": "v2.0",
        "status": "CURRENT",
        "updated": "2012-11-21T11:33:21Z",
        "links": [
            {
                "rel": "describedby",
                "type": "application/pdf",
                "href": "http://jorgew.github.com/block-storage-api/"
                        "content/os-block-storage-1.0.pdf",
            },
            {
                "rel": "describedby",
                "type": "application/vnd.sun.wadl+xml",
                #(anthony) FIXME
                "href": "http://docs.rackspacecloud.com/"
                        "servers/api/v1.1/application.wadl",
            },
        ],
        "media-types": [
            {
                "base": "application/xml",
                "type": "application/vnd.openstack.storage+xml;version=1",
            },
            {
                "base": "application/json",
                "type": "application/vnd.openstack.storage+json;version=1",
            }
        ],
    },
    "v1.0": {
        "id": "v1.0",
        "status": "CURRENT",
        "updated": "2012-01-04T11:33:21Z",
        "links": [
            {
                "rel": "describedby",
                "type": "application/pdf",
                "href": "http://jorgew.github.com/block-storage-api/"
                        "content/os-block-storage-1.0.pdf",
            },
            {
                "rel": "describedby",
                "type": "application/vnd.sun.wadl+xml",
                #(anthony) FIXME
                "href": "http://docs.rackspacecloud.com/"
                        "servers/api/v1.1/application.wadl",
            },
        ],
        "media-types": [
            {
                "base": "application/xml",
                "type": "application/vnd.openstack.storage+xml;version=1",
            },
            {
                "base": "application/json",
                "type": "application/vnd.openstack.storage+json;version=1",
            }
        ],
    }

}

class MediaTypesTemplateElement(xmlutil.TemplateElement):
    def will_render(self, datum):
        return 'media-types' in datum

def make_version(elem):
    elem.set('id')
    elem.set('status')
    elem.set('updated')

    mts = MediaTypesTemplateElement('media-types')
    elem.append(mts)

    mt = xmlutil.SubTemplateElement(mts, 'media-type', selector='media-types')
    mt.set('base')
    mt.set('type')

    xmlutil.make_links(elem, 'links')

version_nsmap = {None: xmlutil.XMLNS_COMMON_V10, 'atom': xmlutil.XMLNS_ATOM}

class VersionTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('version', selector='version')
        make_version(root)
        return xmlutil.MasterTemplate(root, 1, nsmap=version_nsmap)

class VersionsTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('versions')
        elem = xmlutil.SubTemplateElement(root, 'version', selector='versions')
        make_version(elem)
        return xmlutil.MasterTemplate(root, 1, nsmap=version_nsmap)

class ChoicesTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('choices')
        elem = xmlutil.SubTemplateElement(root, 'version', selector='choices')
        make_version(elem)
        return xmlutil.MasterTemplate(root, 1, nsmap=version_nsmap)

class AtomSerializer(wsgi.XMLDictSerializer):

    NSMAP = {None: xmlutil.XMLNS_ATOM}

    def __init__(self, metadata=None, xmlns=None):
        self.metadata = metadata or {}
        if not xmlns:
            self.xmlns = wsgi.XMLNS_ATOM
        else:
            self.xmlns = xmlns

    def _get_most_recent_update(self, versions):
        recent = None
        for version in versions:
            updated = datetime.datetime.strptime(version['updated'],
                                                 '%Y-%m-%dT%H:%M:%SZ')
            if not recent:
                recent = updated
            elif updated > recent:
                recent = updated

        return recent.strftime('%Y-%m-%dT%H:%M:%SZ')

    def _get_base_url(self, link_href):
        # Make sure no trailing /
        link_href = link_href.rstrip('/')
        return link_href.rsplit('/', 1)[0] + '/'

    def _create_feed(self, versions, feed_title, feed_id):
        feed = etree.Element('feed', nsmap=self.NSMAP)
        title = etree.SubElement(feed, 'title')
        title.set('type', 'text')
        title.text = feed_title

        # Set this updated to the most recently updated version
        recent = self._get_most_recent_update(versions)
        etree.SubElement(feed, 'updated').text = recent

        etree.SubElement(feed, 'id').text = feed_id

        link = etree.SubElement(feed, 'link')
        link.set('rel', 'self')
        link.set('href', feed_id)

        author = etree.SubElement(feed, 'author')
        etree.SubElement(author, 'name').text = 'Rackspace'
        etree.SubElement(author, 'uri').text = 'http://www.rackspace.com/'

        for version in versions:
            feed.append(self._create_version_entry(version))

        return feed

    def _create_version_entry(self, version):
        entry = etree.Element('entry')
        etree.SubElement(entry, 'id').text = version['links'][0]['href']
        title = etree.SubElement(entry, 'title')
        title.set('type', 'text')
        title.text = 'Version %s' % version['id']
        etree.SubElement(entry, 'updated').text = version['updated']

        for link in version['links']:
            link_elem = etree.SubElement(entry, 'link')
            link_elem.set('rel', link['rel'])
            link_elem.set('href', link['href'])
            if 'type' in link:
                link_elem.set('type', link['type'])

        content = etree.SubElement(entry, 'content')
        content.set('type', 'text')
        content.text = 'Version %s %s (%s)' % (version['id'],
                                               version['status'],
                                               version['updated'])
        return entry

class VersionsAtomSerializer(AtomSerializer):
    def default(self, data):
        versions = data['versions']
        feed_id = self._get_base_url(versions[0]['links'][0]['href'])
        feed = self._create_feed(versions, 'Available API Versions', feed_id)
        return self._to_xml(feed)

class VersionAtomSerializer(AtomSerializer):
    def default(self, data):
        version = data['version']
        feed_id = version['links'][0]['href']
        feed = self._create_feed([version], 'About This Version', feed_id)
        return self._to_xml(feed)

class Versions(wsgi.Resource):

    def __init__(self):
        super(Versions, self).__init__(None)

    @wsgi.serializers(xml=VersionsTemplate,
                      atom=VersionsAtomSerializer)
    def index(self, req):
        """Return all versions."""
        builder = views_versions.get_view_builder(req)
        return builder.build_versions(VERSIONS)

    @wsgi.serializers(xml=ChoicesTemplate)
    @wsgi.response(300)
    def multi(self, req):
        """Return multiple choices."""
        builder = views_versions.get_view_builder(req)
        return builder.build_choices(VERSIONS, req)

    def get_action_args(self, request_environment):
        """Parse dictionary created by routes library."""
        args = {}
        if request_environment['PATH_INFO'] == '/':
            args['action'] = 'index'
        else:
            args['action'] = 'multi'

        return args

class HardwareVersionV1(object):
    @wsgi.serializers(xml=VersionTemplate,
                      atom=VersionAtomSerializer)
    def show(self, req):
        builder = views_versions.get_view_builder(req)
        return builder.build_version(VERSIONS['v1.0'])

def create_resource():
    return wsgi.Resource(HardwareVersionV1())
